extends VBoxContainer

const NUMBER_OF_SETTINGS = 8

var sound = true
var vibrate = true
var doubletap = false
var showZeros = false
var countDown = true
var showOutline = true
var hideSafeBlocks = false
var highlightZeros = true
var sensitivity = 10

var soundPos = 0
var soundWidth = 150
var vibratePos = 0
var vibrateWidth = 150
var doubletapPos = 0
var doubletapWidth = 150
var showZerosPos = 0
var showZerosWidth = 150
var countDownPos = 0
var countDownWidth = 150
var showOutlinePos = 0
var showOutlineWidth = 150
var hideSafeBlocksPos = 0
var hideSafeBlocksWidth = 150
var highlightZerosPos = 0
var highlightZerosWidth = 150


func _ready() -> void:
	var data = FileOps.loadSettings()
	sound = data.sound
	vibrate = data.vibrate
	doubletap = data.doubletap
	showZeros = data.showZeros
	countDown = data.countDown
	showOutline = data.showOutline
	hideSafeBlocks = data.hideSafeBlocks
	highlightZeros = data.highlightZeros
	sensitivity = data.sensitivity

	$SliderPos/HSlider.value = sensitivity
	$SliderPos/SliderText.position.y -= $SliderPos/SliderText.get_content_height()
	Singleton.lang = getLang(data.language)
	call_deferred("setPositions")
	updateText()


func setPositions() -> void:
	soundPos = $SoundPos/Sound.global_position.y
	vibratePos = $VibratePos/Vibrate.global_position.y
	doubletapPos = $DoubletapPos/Doubletap.global_position.y
	showZerosPos = $ShowZerosPos/ShowZeros.global_position.y
	countDownPos = $CountDownPos/CountDown.global_position.y
	showOutlinePos = $ShowOutlinePos/ShowOutline.global_position.y
	hideSafeBlocksPos = $HideSafeBlocksPos/HideSafeBlocks.global_position.y
	highlightZerosPos = $HighlightZerosPos/HighlightZeros.global_position.y

	setWidth()
	queue_redraw()


func setWidth() -> void:
	soundWidth = $SoundPos/Sound/RichTextLabel.get_content_width()
	vibrateWidth = $VibratePos/Vibrate/RichTextLabel.get_content_width()
	doubletapWidth = $DoubletapPos/Doubletap/RichTextLabel.get_content_width()
	showZerosWidth = $ShowZerosPos/ShowZeros/RichTextLabel.get_content_width()
	countDownWidth = $CountDownPos/CountDown/RichTextLabel.get_content_width()
	showOutlineWidth = $ShowOutlinePos/ShowOutline/RichTextLabel.get_content_width()
	hideSafeBlocksWidth = $HideSafeBlocksPos/HideSafeBlocks/RichTextLabel.get_content_width()
	highlightZerosWidth = $HighlightZerosPos/HighlightZeros/RichTextLabel.get_content_width()


func settingByIdx(idx: int) -> bool:
	return [
		sound, 
		vibrate, 
		doubletap, 
		showZeros, 
		countDown, 
		showOutline, 
		hideSafeBlocks, 
		highlightZeros
	][idx]


func posByIdx(idx: int) -> float:
	return [
		soundPos, 
		vibratePos, 
		doubletapPos, 
		showZerosPos, 
		countDownPos, 
		showOutlinePos, 
		hideSafeBlocksPos,
		highlightZerosPos
	][idx]


func widthByIdx(idx: int) -> float:
	return [
		soundWidth, 
		vibrateWidth, 
		doubletapWidth, 
		showZerosWidth, 
		countDownWidth, 
		showOutlineWidth, 
		hideSafeBlocksWidth,
		highlightZerosWidth
	][idx]


func _draw() -> void:
	for idx in range(NUMBER_OF_SETTINGS):
		var pos = posByIdx(idx)
		var width = widthByIdx(idx)
		var screenWidth = get_viewport().size.x
		var from = -(width / 2) - 20 + screenWidth / 2
		var to = width / 2 + 20 + screenWidth / 2
		draw_line(Vector2(from, pos), Vector2(to, pos), Color.BLACK, 140)
		draw_circle(Vector2(from, pos), 70, Color.BLACK)
		draw_circle(Vector2(to, pos), 70, Color.BLACK)

		if settingByIdx(idx):
			draw_line(Vector2(from, pos), Vector2(to, pos), Color.WHITE, 120)
			draw_circle(Vector2(from, pos), 60, Color.WHITE)
			draw_circle(Vector2(to, pos), 60, Color.WHITE)
		else:
			draw_line(Vector2(from, pos), Vector2(to, pos), Color.DARK_GRAY, 120)
			draw_circle(Vector2(from, pos), 60, Color.DARK_GRAY)
			draw_circle(Vector2(to, pos), 60, Color.DARK_GRAY)


func saveSettings() -> void:
	FileOps.saveSettings(
		sound,
		vibrate,
		sensitivity,
		doubletap,
		showZeros,
		countDown,
		showOutline,
		hideSafeBlocks,
		highlightZeros
	)


func getLang(langNum) -> String:
	match langNum:
		1:
			return "de"
		2:
			return "esp"
		_:
			return "en"


func updateText() -> void:
	$SoundPos/Sound/RichTextLabel.text 	   =  tr("SOUND")
	$VibratePos/Vibrate/RichTextLabel.text 	   =  tr("VIBRATION")
	$DoubletapPos/Doubletap/RichTextLabel.text =  tr("DOUBLETAP")
	$ShowZerosPos/ShowZeros/RichTextLabel.text =  tr("SHOWZEROS")
	$CountDownPos/CountDown/RichTextLabel.text =  tr("COUNTDOWN")
	$ShowOutlinePos/ShowOutline/RichTextLabel.text = tr("SHOWOUTLINE")
	$HideSafeBlocksPos/HideSafeBlocks/RichTextLabel.text = tr("HIDESAFEBLOCKS")
	$HighlightZerosPos/HighlightZeros/RichTextLabel.text = tr("HIGHLIGHTZEROS")
	$SliderPos/SliderText.text = tr("WIGGLEROOM")

	setWidth()


func _on_config_toggle_sound() -> void:
	sound = !sound
	saveSettings()
	queue_redraw()


func _on_config_toggle_vibrate() -> void:
	vibrate = !vibrate
	saveSettings()
	queue_redraw()


func _on_config_reset_text() -> void:
	updateText()
	saveSettings()
	queue_redraw()


func _on_h_slider_drag_ended(value_changed) -> void:
	if value_changed:
		sensitivity = $SliderPos/HSlider.value
		saveSettings()


func _on_config_toggle_count_down() -> void:
	countDown = !countDown
	saveSettings()
	queue_redraw()


func _on_config_toggle_doubletap() -> void:
	doubletap = !doubletap
	saveSettings()
	queue_redraw()


func _on_config_toggle_show_zeros() -> void:
	showZeros = !showZeros
	saveSettings()
	queue_redraw()


func _on_config_toggle_show_outline() -> void:
	showOutline = !showOutline
	saveSettings()
	queue_redraw()


func _on_config_toggle_hide_safe_blocks() -> void:
	hideSafeBlocks = !hideSafeBlocks
	saveSettings()
	queue_redraw()


func _on_config_toggle_highlight_zeros() -> void:
	highlightZeros = !highlightZeros
	saveSettings()
	queue_redraw()
