extends Node3D

const MAX_UNDO_STEPS = 50

var Box   = preload("res://scenes/Box.tscn")
var Ghost = preload("res://scenes/Ghost.tscn")

var size  = 9
var shape = "cube"
var showOutline  = true
var activeBlocks = 0
var activeMines  = 0
var totalMines   = 0
var nonMines     = []
var mines		 = []
var gameTime     = 0.0
var isLoading	 = false
var keepSaving   = true		# to disable saving when all remaining mines are marked

var undoStack	 = []

signal exploded
signal mineCount(count)
signal won(time)
signal setUndo(state)


func _process(delta):
	if !Singleton.first_press:
		gameTime += delta


func _ready():

	size = Singleton.fieldSize

	position = Vector3i(float(size) / 2, float(size) / 2, float(size) / 2)

	var hs = int(size / 2)
	var mod = 0

	if hs * 2 < size:
		mod = 1

	if Singleton.loadGame && FileOps.saveExists(Singleton.page):
		loadGameBinary()

	else:
		if Singleton.page == 2: # daily challenge
			setDailySeed()
			FileOps.lockIn(size)

		Singleton.first_press = true
		var field = []

		var dummy = Box.instantiate()
		dummy.triggered = true
		var dummyRow = []
		var dummySlice = []

		for x in range(0, size + 2):
			dummyRow.append(dummy)
		for x in range(0, size + 2):
			dummySlice.append(dummyRow)

		if shape == "cube":
			activeBlocks = size ** 3
			field.append(dummySlice)
			for x in range(-hs, hs + mod):
				var slice = []
				slice.append(dummyRow)
				for y in range(-hs, hs + mod):
					var row = []
					row.append(dummy)
					for z in range(-hs, hs + mod):

						var box = Box.instantiate()
						box.exploded.connect(die)
						box.cleared.connect(clearBox)
						box.defused.connect(markMine)
						box.refused.connect(rearmMine)
						box.passMine.connect(passMine)
						box.clearUndoStack.connect(clearUndoStack)

						var rand = randi_range(1, 10)

						box.position = Vector3i(x, y, z)

						if rand == 5:
							box.mine = true
							activeMines += 1
							mines.append(box)
						else:
							nonMines.append(box)

						row.append(box)
						

					row.append(dummy)
					slice.append(row)
				slice.append(dummyRow)

				field.append(slice)
			field.append(dummySlice)
		totalMines = activeMines

		var id = 0
		for z in range(0, size + 2):
			for y in range(0, size + 2):
				for x in range(0, size + 2):
					if x == 0 or x == size + 1 or y == 0 or y == size + 1 or z == 0 or z == size + 1:
						pass
					else:
						field[z][y][x].neighbors = [
							field[z][y][x + 1],
							field[z][y][x - 1],

							field[z][y + 1][x + 1],
							field[z][y + 1][x - 1],
							field[z][y + 1][x],
							field[z][y - 1][x + 1],
							field[z][y - 1][x],
							field[z][y - 1][x - 1], 

							field[z + 1][y][x + 1],
							field[z + 1][y][x],
							field[z + 1][y][x - 1],
							field[z + 1][y + 1][x + 1],
							field[z + 1][y + 1][x - 1],
							field[z + 1][y + 1][x],
							field[z + 1][y - 1][x + 1],
							field[z + 1][y - 1][x],
							field[z + 1][y - 1][x - 1], 

							field[z - 1][y][x + 1],
							field[z - 1][y][x],
							field[z - 1][y][x - 1],
							field[z - 1][y + 1][x + 1],
							field[z - 1][y + 1][x - 1],
							field[z - 1][y + 1][x],
							field[z - 1][y - 1][x + 1],
							field[z - 1][y - 1][x],
							field[z - 1][y - 1][x - 1]
						].filter(
							func(realBoxes): return !realBoxes.triggered
						)

						field[z][y][x].id = id
						add_child(field[z][y][x])
						id += 1
		reCenter()

	if showOutline:
		var ghost = Ghost.instantiate()
		ghost.scale = Vector3i(size, size, size)
		add_child(ghost)

	updateMines()
	checkState()


# seed for daily challenges
func setDailySeed():
	seed(
		hash(
			str(Time.get_date_dict_from_system()) +
			"Bombardiro_Crocodilo" # salt
		)
	)


func reCenter():

	var center = size / 2
	var centerPoint = Vector3(center, center, center)
	var upVec = Vector3(0, 1, 0)

	var sphere = get_parent().get_node("PlayerSphere")
	var cam    = get_parent().get_node("PlayerSphere/Cam")

	sphere.position = centerPoint
	
	if size == 5:
		cam.position = Vector3(1.5, 5, 10)
	elif size == 9:
		cam.position = Vector3(2.5, 7.5, 14)
	else:
		cam.position = Vector3(4, 10, 18)

	cam.global_position = cam.position + Vector3(
		int(center),
		int(center),
		int(center)
	)
	cam.look_at(centerPoint, upVec)


func die():
	keepSaving = false
	FileOps.deleteSave(Singleton.page)
	exploded.emit()
	var data = FileOps.loadStats(Singleton.page)
	data.totalMines += totalMines - activeMines
	data.playedGames += 1
	if size == 5:
		data.totalEasy += 1
	elif size == 9:
		data.totalMedium += 1
	else:
		data.totalHard += 1

	FileOps.saveStats(data, Singleton.page)

	if Singleton.page == 2:
		var dailyData = FileOps.loadDaily()
		
		if size == 5:
			dailyData.playedBeginner = true
		elif size == 9:
			dailyData.playedIntermediate = true
		else:
			dailyData.playerMaster = true

		FileOps.saveDaily(dailyData)


func clearBox():
	activeBlocks -= 1
	checkState()


func markMine():
	activeMines -= 1
	checkState()
	updateMines()


func rearmMine():
	activeMines += 1
	updateMines()


func updateMines():
	mineCount.emit(activeMines)

	if keepSaving:
		saveGameBinary()


func checkState():
	if activeBlocks == totalMines and keepSaving: # and activeMines == 0: may need a toggle
		FileOps.deleteSave(Singleton.page)
		keepSaving = false
		var totalTime = round(gameTime * pow(10.0, 2)) / pow(10.0, 2)
		won.emit(totalTime)
		for box in mines:
			if not box.marked:
				box.mark()
				await get_tree().create_timer(0.01).timeout

		var data = FileOps.loadStats(Singleton.page)
		data.totalMines += totalMines
		data.playedGames += 1
		data.wonGames += 1
		if size == 5:
			data.totalEasy += 1
			data.wonEasy += 1
			if data.fastestEasyMines == -1 or data.fastestEasy > totalTime:
				data.fastestEasy = totalTime
				data.fastestEasyMines = totalMines
		elif size == 9:
			data.totalMedium += 1
			data.wonMedium += 1
			if data.fastestMediumMines == -1 or data.fastestMedium > totalTime:
				data.fastestMedium = totalTime
				data.fastestMediumMines = totalMines
		else:
			data.totalHard += 1
			data.wonHard += 1
			if data.fastestHardMines == -1 or data.fastestHard > totalTime:
				data.fastestHard = totalTime
				data.fastestHardMines = totalMines

		FileOps.saveStats(data, Singleton.page)

		if Singleton.page == 2:
			var dailyData = FileOps.loadDaily()

			if size == 5:
				dailyData.playedBeginner = true
				dailyData.timeBeginner = totalTime
			elif size == 9:
				dailyData.playedIntermediate = true
				dailyData.timeIntermediate = totalTime
			else:
				dailyData.playerMaster = true
				dailyData.timeMaster = totalTime

			FileOps.saveDaily(dailyData)


func passMine():
	nonMines.pick_random().recieve_mine()


func addToUndoStack(box, mark: bool):
	if undoStack.is_empty():
		setUndo.emit(true)

	undoStack.append([weakref(box), mark])

	if undoStack.size() > MAX_UNDO_STEPS:
		undoStack.pop_front()


func undo():
	if !undoStack.is_empty():
		var culprit = undoStack.pop_back()
		var box = culprit[0].get_ref()
		
		if culprit[1]:
			box.unmark(false)
		else:
			box.mark(false)

		if undoStack.is_empty():
			setUndo.emit(false)


func clearUndoStack():
	undoStack.clear()
	setUndo.emit(false)


func saveAndExit():
	saveGameBinary()
	get_tree().change_scene_to_file("res://scenes/Settings.tscn")


func saveGameBinary():

	if !keepSaving:
		return

	var file = FileAccess.open(FileOps.getSaveForPage(Singleton.page), FileAccess.WRITE)
	if file == null:
		print("Error opening file for writing: ", FileAccess.get_open_error())
		return

	# Save WorldData
	var world = get_parent()
	var sphere = world.get_node("PlayerSphere")

	var transform = sphere.transform
	file.store_real(transform.basis.x.x)
	file.store_real(transform.basis.x.y)
	file.store_real(transform.basis.x.z)
	file.store_real(transform.basis.y.x)
	file.store_real(transform.basis.y.y)
	file.store_real(transform.basis.y.z)
	file.store_real(transform.basis.z.x)
	file.store_real(transform.basis.z.y)
	file.store_real(transform.basis.z.z)
	file.store_real(transform.origin.x)
	file.store_real(transform.origin.y)
	file.store_real(transform.origin.z)

	file.store_real(sphere.scale.x)
	file.store_real(sphere.scale.y)
	file.store_real(sphere.scale.z)

	file.store_real(gameTime)
	file.store_8(1 if Singleton.first_press else 0)

	file.store_32(size) # Store the field size as a 32-bit integer

	# Save Box data
	var allBoxes = []
	for box in nonMines + mines:
		allBoxes.append(box)
	allBoxes.sort_custom(func(a, b): return a.id < b.id)

	file.store_32(allBoxes.size())

	for box in allBoxes:
		# Pack boolean states into a single byte
		var byteData = 0
		if box.mine: byteData |= (1 << 0)
		if box.triggered: byteData |= (1 << 1)
		if box.marked: byteData |= (1 << 2)
		if box.active: byteData |= (1 << 3)
		file.store_8(byteData)

		# Store initialMineCount
		file.store_8(box.initialMineCount)

		# Store position
		var gridPosX = box.position.x + float(size) / 2.0
		var gridPosY = box.position.y + float(size) / 2.0
		var gridPosZ = box.position.z + float(size) / 2.0

		file.store_32(int(gridPosX)) # Store as int
		file.store_32(int(gridPosY))
		file.store_32(int(gridPosZ))

		# Store number of neighbors and their IDs
		file.store_32(box.neighbors.size())
		for neighbor in box.neighbors:
			file.store_32(neighbor.id)
			
	file.close()


func loadGameBinary():

	if not FileOps.saveExists(Singleton.page):
		print("Save file not found: ", FileOps.getSaveForPage(Singleton.page))
		return

	isLoading = true
	var file = FileAccess.open(FileOps.getSaveForPage(Singleton.page), FileAccess.READ)
	if file == null:
		print("Error opening file for reading: ", FileAccess.get_open_error())
		isLoading = false
		return

	# Load WorldData
	var transformBasisX = Vector3(file.get_real(), file.get_real(), file.get_real())
	var transformBasisY = Vector3(file.get_real(), file.get_real(), file.get_real())
	var transformBasisZ = Vector3(file.get_real(), file.get_real(), file.get_real())
	var transformOrigin = Vector3(file.get_real(), file.get_real(), file.get_real())
	var loadedTransform = Transform3D(transformBasisX, transformBasisY, transformBasisZ, transformOrigin)

	var loadedZoom = Vector3(file.get_real(), file.get_real(), file.get_real())
	gameTime = file.get_real()
	Singleton.first_press = file.get_8() == 1

	size = file.get_32() # Load the field size
	Singleton.fieldSize = size

	self.position = Vector3i(float(size) / 2, float(size) / 2, float(size) / 2)
	reCenter()

	var world = get_parent()
	world.get_node("PlayerSphere").transform = loadedTransform
	world.get_node("PlayerSphere").scale = loadedZoom

	# Load Box data
	activeBlocks = 0
	activeMines  = 0
	totalMines   = 0
	mines.clear()
	nonMines.clear()

	var loadedBoxesMap = {} # Map to reference boxes by ID

	var numBoxes = file.get_32()

	for i in range(numBoxes):
		var box = Box.instantiate()
		box.id = i # ID corresponds to the index

		var byteData = file.get_8()
		box.mine = (byteData & (1 << 0)) != 0
		box.triggered = (byteData & (1 << 1)) != 0
		box.marked = (byteData & (1 << 2)) != 0
		box.active = (byteData & (1 << 3)) != 0
		
		box.initialMineCount = file.get_8()

		var loadedBoxPosX = ceil(file.get_32() - float(size) / 2.0)
		var loadedBoxPosY = ceil(file.get_32() - float(size) / 2.0)
		var loadedBoxPosZ = ceil(file.get_32() - float(size) / 2.0)

		box.position = Vector3(loadedBoxPosX, loadedBoxPosY, loadedBoxPosZ)
   
		var numNeighbors = file.get_32()
		box.neighborsIds.clear()
		for j in range(numNeighbors):
			box.neighborsIds.append(file.get_32())

		if box.mine:
			activeMines += 1
			totalMines += 1
			mines.append(box)
		else:
			nonMines.append(box)

		if not box.triggered and not box.marked:
			activeBlocks += 1

		loadedBoxesMap[box.id] = box

	file.close()

	var markedBoxes = []
	# Build neighbor relationships and initialize boxes
	for key in loadedBoxesMap.keys():
		var box = loadedBoxesMap[key]
		for id in box.neighborsIds:
			if loadedBoxesMap.has(id):
				box.neighbors.append(loadedBoxesMap[id])

		# Initialize the box after neighbor relationships are built
		box.exploded.connect(die)
		box.cleared.connect(clearBox)
		box.defused.connect(markMine)
		box.refused.connect(rearmMine)
		box.passMine.connect(passMine)
		box.clearUndoStack.connect(clearUndoStack)

		add_child(box)

		if box.triggered:
			box.reTrigger()

		if box.marked:
			box.turnOrange()
			activeMines -= 1
			activeBlocks += 1
			markedBoxes.append(box)

		box.refreshMineCount()
		box.update()

	if Singleton.highlightZeros:
		for markedBox in markedBoxes:
			markedBox.reColor()

	isLoading = false
