extends WorldEnvironment


signal undoPress
signal saveGame

var MineField = preload("res://scenes/MineField.tscn")
var zooming = false
var over = false
var reallyOver = false
var reallyReallyOver = false # so that doubletab doesn't skip the ending screens
var pressPoint = Vector2.ZERO

func _ready():
	get_tree().quit_on_go_back = false
	genesis()


func genesis():

	var minefield = MineField.instantiate()

	minefield.size = Singleton.fieldSize
	minefield.shape = "cube"
	minefield.showOutline = Singleton.showOutline

	minefield.exploded.connect(dead)
	minefield.mineCount.connect(updateCount)
	minefield.won.connect(win)
	minefield.setUndo.connect(setUndo)

	undoPress.connect(minefield.undo)
	saveGame.connect(minefield.saveAndExit)

	add_child(minefield)


func _notification(what):
	match what:
		NOTIFICATION_WM_GO_BACK_REQUEST:
			get_tree().change_scene_to_file("res://scenes/Settings.tscn")


func _input(event):
	var delta = 0.005

	var is_drag = event is InputEventScreenDrag or event is InputEventMouseMotion
	
	var is_press = false
	var unlock_rotation = false
	if event is InputEventScreenTouch:
		is_press = !is_drag and event.double_tap if Singleton.doubletap else event.is_pressed()
		unlock_rotation = !is_drag and !Singleton.dragging and event.is_pressed()

	if event is InputEventMagnifyGesture:

		Singleton.dragging = true
		zooming = true

		if reallyOver:
			reallyOver = false

		$PlayerSphere.scale *= (event.factor - 2) * -1
		$PlayerSphere.scale = clamp(
			$PlayerSphere.scale, 
			Vector3(.1, .1, .1), 
			Vector3(5, 5, 5)
		)


	elif event is InputEventPanGesture:
		Singleton.dragging = true
		zooming = true
		$PlayerSphere.rotate_object_local(Vector3.FORWARD, event.delta.x * delta)


	if event is InputEventScreenDrag and !zooming:
		if pressPoint.distance_to(event.position) > Singleton.sensitivity:
			Singleton.dragging = true

		if reallyOver:
			reallyOver = false

		var cam = $PlayerSphere/Cam
		var right_axis = -cam.global_transform.basis.x.normalized()
		var up_axis = cam.global_transform.basis.y.normalized()
		
		$PlayerSphere.rotate(right_axis, event.relative.y * delta)
		$PlayerSphere.rotate(up_axis, -event.relative.x * delta)

	if event is InputEventScreenTouch:
		if unlock_rotation:
			zooming = false
		
		if is_press:
			pressPoint = event.position
			zooming = false
			if over:
				reallyOver = true

			if reallyReallyOver and Singleton.doubletap:
				if Singleton.page == 2:
					get_tree().change_scene_to_file("res://scenes/Settings.tscn")
				else:
					get_tree().change_scene_to_file("res://scenes/LoadingScreenLight.tscn") 

		else:
			if reallyOver and !Singleton.doubletap:
				if Singleton.page == 2:
					get_tree().change_scene_to_file("res://scenes/Settings.tscn")
				else:
					get_tree().change_scene_to_file("res://scenes/LoadingScreenLight.tscn")
			elif reallyOver:
				reallyReallyOver = true


func dead():
	$PlayerSphere/Cam/Menu/UI/Text.text =  "[center]" + tr("EXPLODED") + "[/center]"
	$PlayerSphere/Cam/Menu/UI.show()
	over = true


func updateCount(count):
	$PlayerSphere/Cam/UI/UILayer/MinesCount.text = "[center] " + str(count) + " "


func win(time):
	$PlayerSphere/Cam/Menu/UI/Text.text = "[center]" + tr("WON_1") + str(time) + tr("WON_2") + "[/center]"
	$PlayerSphere/Cam/Menu/UI/Text.add_theme_color_override("default_color", Color("green"))
	$PlayerSphere/Cam/Menu/UI.show()
	over = true


func setUndo(state):
	var label = $PlayerSphere/Cam/UndoButton/RichTextLabel
	if state:
		label.text = "[color=#ffffff][b]↩[/b][/color]"  # White when active
	else:
		label.text = "[color=#777777][b]↩[/b][/color]"  # Gray when inactive


func _on_undo() -> void:
	undoPress.emit()


func _on_back_save_game() -> void:
	saveGame.emit()
